﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Umgebungsbedingungssensors BME280 Bosch                                             //
//  ________________________________________________________________________            //
//                                                                                      //
//  Copyright (c) 2019 Reimar Grasbon												    //
//                                                                                      //
//  Grundlage der Software sind die Kompensationsmethoden aus dem BME280-Datenblatt und //
//  des BME280-API der Bosch Sensortec GmbH, erhältlich unter www.bosch-sensortec.com.  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include "Conf.h"
#include "Board.h"
#include "Timer.h"
#include "BME280.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef struct
{   UInt16  digT1;
    SInt16  digT2;
    SInt16  digT3;
    UInt16  digP1;
    SInt16  digP2;
    SInt16  digP3;
    SInt16  digP4;
    SInt16  digP5;
    SInt16  digP6;
    SInt16  digP7;
    SInt16  digP8;
    SInt16  digP9;
    UInt08  digH1;
    SInt16  digH2;
    UInt08  digH3;
    SInt16  digH4;
    SInt16  digH5;
    SInt08  digH6;
    SInt32  tFine;
} CalibrationData;

typedef struct
{   UInt32 pressure;    // Unkompensierter Druckwert
    UInt32 temperature; // Unkompensierter Temperaturwert
    UInt32 humidity;    // Unkompensierter Feuchtewert
} RawData;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static  bool             isConnected = false;
static  CalibrationData  calibrationData;
static  AmbientConditions       conditions = { INVALID, INVALID, INVALID };

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static bool
Reset(void)
{
    UInt08 data[] = { 0xE0, 0xB6 };     // Soft-Reset: Wert 0xB6 -> Register E0
    return BOARD_I2cTransmit(CFG_I2C_ADR_BME280, data, CARDINALITY(data));
}

//////////////////////////////////////////////////////////////////////////////////////////
static bool
Initialize(void)
{
    // Kalibrierungsdaten-Teil 1, Register 88 .. A1 "calib00 .. calib26"
    {   UInt08 regAddr[1] = { 0x88 };
        if (!BOARD_I2cTransmit(CFG_I2C_ADR_BME280, regAddr, CARDINALITY(regAddr))) return false;
        UInt08 regData[26];
        if (!BOARD_I2cReceive(CFG_I2C_ADR_BME280, regData, CARDINALITY(regData))) return false;
        calibrationData.digT1 = ToWord(regData[ 1], regData[ 0]);
        calibrationData.digT2 = ToWord(regData[ 3], regData[ 2]);
        calibrationData.digT3 = ToWord(regData[ 5], regData[ 4]);
        calibrationData.digP1 = ToWord(regData[ 7], regData[ 6]);
        calibrationData.digP2 = ToWord(regData[ 9], regData[ 8]);
        calibrationData.digP3 = ToWord(regData[11], regData[10]);
        calibrationData.digP4 = ToWord(regData[13], regData[12]);
        calibrationData.digP5 = ToWord(regData[15], regData[14]);
        calibrationData.digP6 = ToWord(regData[17], regData[16]);
        calibrationData.digP7 = ToWord(regData[19], regData[18]);
        calibrationData.digP8 = ToWord(regData[21], regData[20]);
        calibrationData.digP9 = ToWord(regData[23], regData[22]);
        calibrationData.digH1 = regData[25];
    }
    // Kalibrierungsdaten-Teil 2, Register E1 .. E7 "calib26 .. calib32"
    {   UInt08 regAddr[1] = { 0xE1 };
        if (!BOARD_I2cTransmit(CFG_I2C_ADR_BME280, regAddr, CARDINALITY(regAddr))) return false;
        UInt08 regData[7];
        if (!BOARD_I2cReceive(CFG_I2C_ADR_BME280, regData, CARDINALITY(regData))) return false;
        calibrationData.digH2 = ToWord(regData[1], regData[0]);
        calibrationData.digH3 = regData[2];
        SInt16 digH4Msb = (SInt16)(SInt08)regData[3] << 4;  //    E4          E5
        SInt16 digH4Lsb = (SInt16)(regData[4] & 0x0F);      //  xxxxxxxx   ····xxxx
        calibrationData.digH4 = digH4Msb | digH4Lsb;        //
        SInt16 digH5Msb = (SInt16)(SInt08)regData[5] << 4;  //    E5          E6
        SInt16 digH5Lsb = (SInt16)(regData[4] >> 4);        //  xxxx····   xxxxxxxx
        calibrationData.digH5 = digH5Msb | digH5Lsb;        //
        calibrationData.digH6 = (SInt08)regData[6];
    }
    // Konfiguration (Register F5: "config"):
    {   //                        ___________ t_sb:     Pause zwischen Messungen: 125 ms
        //                        |||________ filter:   IIR-Filterbreite: 4
        //                        |||||| ____ spi3w_en: 3-Draht-SPI Schnittstelle: aus
        //                        |||||| |
        //                      0b01010000
        UInt08 data[] = { 0xF5, 0x50 };
        if (!BOARD_I2cTransmit(CFG_I2C_ADR_BME280, data, CARDINALITY(data))) return false;
    }
    // Einstellung Feuchte (Register F2: "ctrl_hum"):
    {   //                             ______ osrs_h: Feuchte-Überabtastung: 4x
        //                             |||
        //                      0b00000010
        UInt08 data[] = { 0xF2, 0x02 };
        if (!BOARD_I2cTransmit(CFG_I2C_ADR_BME280, data, CARDINALITY(data))) return false;
    }
    // Einstellung Temperatur, Druck und Sensor-Modus (Register F4 "ctrl_meas"):
    {   //                        __________ osrs_t: Temperatur-Überabtastung: 4x
        //                        |||_______ osrs_p: Druck-Überabtastung: 4x
        //                        ||||||____ model:  Sensor-Betriebsart: Normaler Modus
        //                        ||||||||
        //                      0b01101111
        UInt08 data[] = { 0xF4, 0x6F };
        if (!BOARD_I2cTransmit(CFG_I2C_ADR_BME280, data, CARDINALITY(data))) return false;
    }
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
static bool
ReadRawData(RawData* rawData)
{
    // Register F7 .. FE "press_* temp_* hum_*"
    UInt08 regAddr[1] = { 0xF7 };
    if (!BOARD_I2cTransmit(CFG_I2C_ADR_BME280, regAddr, CARDINALITY(regAddr))) return false;
    UInt08 regData[8];
    if (!BOARD_I2cReceive(CFG_I2C_ADR_BME280, regData, CARDINALITY(regData))) return false;
    UInt32 data_xlsb;
    UInt32 data_lsb;
    UInt32 data_msb;
    // Druck-Rohwert
    data_msb  = (UInt32)regData[0] << 12;
    data_lsb  = (UInt32)regData[1] << 4;
    data_xlsb = (UInt32)regData[2] >> 4;
    rawData->pressure = data_msb | data_lsb | data_xlsb;
    // Temperatur-Rohwert
    data_msb  = (UInt32)regData[3] << 12;
    data_lsb  = (UInt32)regData[4] << 4;
    data_xlsb = (UInt32)regData[5] >> 4;
    rawData->temperature = data_msb | data_lsb | data_xlsb;
    // Feuchte-Rohwert
    data_lsb = (UInt32)regData[6] << 8;
    data_msb = (UInt32)regData[7];
    rawData->humidity = data_msb | data_lsb;
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
static SInt16
CalculateTemperature(UInt32 rawTemperature)
{
    SInt32 var1;
    var1 = (SInt32)((rawTemperature / 8) - ((SInt32)calibrationData.digT1 * 2));
    var1 = (var1 * ((SInt32)calibrationData.digT2)) / 2048;
    SInt32 var2;
    var2 = (SInt32)((rawTemperature / 16) - ((SInt32)calibrationData.digT1));
    var2 = (((var2 * var2) / 4096) * ((SInt32)calibrationData.digT3)) / 16384;
    calibrationData.tFine = var1 + var2;
    SInt32 temperature = (calibrationData.tFine * 5 + 128) / 256; // [°C÷100]
    if      (temperature < -5000) return INVALID;
    else if (temperature >  9900) return INVALID;
    else                          return (SInt16)((temperature+50)/10);   // [°C÷10]
}

//////////////////////////////////////////////////////////////////////////////////////////
static SInt16
CalculatePressure(UInt32 rawPressure)
{
    SInt32 var1 = (((SInt32)calibrationData.tFine) / 2) - (SInt32)64000;
    SInt32 var2 = (((var1 / 4) * (var1 / 4)) / 2048) * ((SInt32)calibrationData.digP6);
    var2 = var2 + ((var1 * ((SInt32)calibrationData.digP5)) * 2);
    var2 = (var2 / 4) + (((SInt32)calibrationData.digP4) * 65536);
    SInt32 var3 = (calibrationData.digP3 * (((var1 / 4) * (var1 / 4)) / 8192)) / 8;
    SInt32 var4 = (((SInt32)calibrationData.digP2) * var1) / 2;
    var1 = (var3 + var4) / 262144;
    var1 = (((32768 + var1)) * ((SInt32)calibrationData.digP1)) / 32768;
    if (var1 == 0) return INVALID; // Division durch Null Vorbeugung
    UInt32 var5 = (UInt32)((UInt32)1048576) - rawPressure;
    UInt32 pressure = ((UInt32)(var5 - (UInt32)(var2 / 4096))) * 3125; // [hPa÷100]
    if (pressure < 0x80000000) pressure = (pressure << 1) / ((UInt32)var1);
    else                       pressure = (pressure / (UInt32)var1) * 2;
    var1 = (((SInt32)calibrationData.digP9) * ((SInt32)(((pressure / 8) * (pressure / 8)) / 8192))) / 4096;
    var2 = (((SInt32)(pressure / 4)) * ((SInt32)calibrationData.digP8)) / 8192;
    pressure = (UInt32)((SInt32)pressure + ((var1 + var2 + calibrationData.digP7) / 16));
    if      (pressure <  30000) return INVALID;
    else if (pressure > 110000) return INVALID;
    else                        return (SInt16)((pressure+50)/100); // [hPa]
}

//////////////////////////////////////////////////////////////////////////////////////////
static SInt16
CalculateHumidity(UInt32 rawHumidity)
{
    SInt32 var1 = calibrationData.tFine - ((SInt32)76800);
    SInt32 var2 = (SInt32)(rawHumidity * 16384);
    SInt32 var3 = (SInt32)(((SInt32)calibrationData.digH4) * 1048576);
    SInt32 var4 = ((SInt32)calibrationData.digH5) * var1;
    SInt32 var5 = (((var2 - var3) - var4) + (SInt32)16384) / 32768;
    var2 = (var1 * ((SInt32)calibrationData.digH6)) / 1024;
    var3 = (var1 * ((SInt32)calibrationData.digH3)) / 2048;
    var4 = ((var2 * (var3 + (SInt32)32768)) / 1024) + (SInt32)2097152;
    var2 = ((var4 * ((SInt32)calibrationData.digH2)) + 8192) / 16384;
    var3 = var5 * var2;
    var4 = ((var3 / 32768) * (var3 / 32768)) / 128;
    var5 = var3 - ((var4 * ((SInt32)calibrationData.digH1)) / 16);
    var5 = (var5 < 0 ? 0 : var5);
    var5 = (var5 > 419430400 ? 419430400 : var5);
    UInt32 humidity = (UInt32)(var5 / 4096); // humidity = Q22.10 [%RH]
    if (humidity > 102400) return INVALID;
    else                   return (SInt16)((humidity*10+50)/1024); //  [%RH÷10]
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
bool
BME280_IsConnected(void)
{
	return isConnected;
}

//////////////////////////////////////////////////////////////////////////////////////////
const AmbientConditions*
BME280_Conditions(void)
{
    return &conditions;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BME280_Startup(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BME280_Proceed(void)
{
    typedef enum
    {   RESET,
        INITIALIZE,
        WORK
    } State;

    static State state  = RESET;
    static Timer  timer = TIMER_MILLISECONDS(0);
    if (!TIMER_RestartIfExpired(&timer)) return;
    RawData rawData;
    switch (state)
    {   case RESET:
            isConnected = Reset();
			if (isConnected)
            {   state = INITIALIZE;
                TIMER_Start(&timer, 50);
            }
            else
            {   TIMER_Start(&timer, 1500);  // Neuer Versuch in 1.5 Sekunden
            }
            break;
        case INITIALIZE:
            isConnected = Initialize();
			if (isConnected)
            {   state = WORK;
                TIMER_Start(&timer, 50);
            }
            else
            {   state = RESET;
                TIMER_Start(&timer, 1500);  // Neustart in 1,5 Sekunden
            }
            break;
        case WORK:
            isConnected = ReadRawData(&rawData);
			if (isConnected)
            {   conditions.temperature = CalculateTemperature(rawData.temperature);
                conditions.pressure    = CalculatePressure(rawData.pressure);
                conditions.humidity    = CalculateHumidity(rawData.humidity);
                TIMER_Start(&timer, CFG_BME280_READOUT_INTERVAL*1000);
            }
            else
            {   conditions.pressure    = INVALID;
                conditions.humidity    = INVALID;
                conditions.temperature = INVALID;
				state = RESET;
                TIMER_Start(&timer, 1500);  // Neustart in 1,5 Sekunden
            }
            break;
        default:
            state = RESET;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BME280_Shutdown(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
